--  widgets.e
--  jiri_babor@hotmail.com
--  00-09-04
--  version 2.40

--==================================================================--
--  This version of widgets is a fusion of
--      font.e      version 6.00  (included below)
--      rat.e       version 2.50  (merged: trimmed & modified)
--      list.e      version 2.00  (merged)
--      ports.e     Jacques Deschenes (merged: Output only)
--      Pete's smart_print -  mods by Gabriel Boehme & me (jiri)
--      widgets     core code
--      handful of miscellaneous routines
--==================================================================--


--without warning
--without type_check

include font.e          -- needs graphics.e, machine.e, misc.e
include machine.e

USE_DOS_KEYS = True

----------------------------------------------------------------------
--  misc common declaration
----------------------------------------------------------------------

global constant
    false = 0,
    true = 1,
    enter = 13,
    esc = 27,
    space = 32

global function max(sequence s)     -- too simple, but fast!
    atom a
    a = s[1]
    for i=2 to length(s) do
	if a < s[i] then a = s[i] end if
    end for
    return a
end function

global function abs(object x)      -- Carl's
    return x*((x>0)-(x<0))
end function

global function round(object x)
--  simple, universal & fast
--  it rounds mid-points *up*, both negative and positive

    if atom(x) then
	return floor(x+0.5)
    end if
    for i=1 to length(x) do
	x[i] = round(x[i])
    end for
    return x
end function -- round

global procedure put_image(integer x,integer y,sequence image)
--  almost same as display_image()

    for i=1 to length(image) do
	pixel(image[i], {x,y})
	y += 1
    end for
end procedure -- put_image

global procedure xor_image(integer x,integer y,sequence im)
--  similar to display_image(): 2-d sequence of pixels im is xor-ed with
--  screen (background) starting at point x,y (left top corner)

    for i=1 to length(im) do
	pixel(xor_bits(get_pixel({x,y}&length(im[i])),im[i]),{x,y})
	y += 1
    end for
end procedure -- xor_image

global procedure merge_image(
    integer x,      -- top left corner, horizontal coordinate
    integer y,      -- top left corner, vertical coordinates
    sequence image, -- 2-d image sequence
    integer c       -- transparent cell color
    )
--  display 2-d sequence im at point x,y
--  image cells of color c are considered transparent
--  with 'on-the-fly' run-length-encoding

    sequence row
    integer f,k,w

    x -= 1
    for i=1 to length(image) do
	row=image[i]
	w = length(row)
	f=0                 -- not in run
	for j=1 to w do
	    if row[j]=c then
		if f then
		    pixel(row[k..j-1],{x+k,y})
		    f=0
		end if
	    else
		if not f then
		    k=j
		    f=1
		end if
	    end if
	end for
	if f then
	    pixel(row[k..w],{x+k,y})
	end if
	y+=1
    end for
end procedure -- merge_image

global function replicate(
    sequence pattern,
    integer newlength,  -- required patterned length
    integer offset      -- left edge pattern offset for global tiling
    )
--  return a sequence generated by pattern repetition
--  for *local* patterns the offset is *zero*
--  for *global* patterns the offset is the starting
--  horizontal location in global coordinates

    if offset then
	offset = remainder(offset,length(pattern))
	newlength += offset
    end if
    while length(pattern) < newlength do
	pattern &= pattern
    end while
    return pattern[1+offset..newlength]
end function -- replicate

global function tile(   -- thanks to C. Taylor for his contribution
    sequence pattern,   -- 2d patern, tile
    integer newwidth,   -- width of tiled area
    integer newheight,  -- height of tiled area
    integer xoffset,    -- left edge for global tiling
    integer yoffset     -- top edge for global tiling
    )
--  return 2-d sequence (image) of tiled area
--  for local tiling xoffset and yoffset are zeros, and
--  the pattern will start in the top left corner of tiled area

    if newheight <= length(pattern) then
	pattern = replicate(pattern, newheight, yoffset)
	for i = 1 to newheight do
	    pattern[i] = replicate(pattern[i], newwidth, xoffset)
	end for
	return pattern
    else
	for i = 1 to length(pattern) do
	    pattern[i] = replicate(pattern[i], newwidth, xoffset)
	end for
	return replicate(pattern, newheight, yoffset)
    end if
end function -- tile

global procedure bar(
    object c,       -- color or 2-d pattern (tile)
    integer x,      -- left edge
    integer y,      -- top edge
    integer width,
    integer height
    )
--  draw filled rectangle
    if atom(c) then
	polygon(c,1,{{x,y},{x+width-1,y},{x+width-1,y+height-1},{x,y+height-1}})
    else
	display_image({x,y},tile(c,width,height,x,y))
    end if
end procedure -- bar

global procedure frame(
    object c,       -- if atom(c) then it is the line color,
		    -- else it is frame colors c={{i1,i2,...},{j1,j2...}},
		    -- where i1, i2,... and j1, j2,...  are top left and
		    -- bottom right edges respectively, going from outside
    integer x,      -- left edge
    integer y,      -- top edge
    integer w,      -- width
    integer h       -- height
    )
    if atom(c) then
	polygon(c,0,{{x,y},{x+w-1,y},{x+w-1,y+h-1},{x,y+h-1}})
    else
	for i=1 to length(c[1]) do
	    draw_line(c[1][i], {{x+i-1,y+h-i},{x+i-1,y+i-1},{x+w-i-1,y+i-1}})
	end for
	for i=1 to length(c[2]) do
	    draw_line(c[2][i], {{x+i,y+h-i},{x+w-i,y+h-i},{x+w-i,y+i-1}})
	end for
    end if
end procedure -- frame

global procedure set_colors(integer c, sequence s)
--  set contiguous range of colors
--  c is starting color index
--  s is sequence of {r,g,b} palette entries
--  rgb values range 0..63
integer ok

    for i=1 to length(s) do
	ok = vga_setpalette({c,s[i][1],s[i][2],s[i][3]})
	c+=1
    end for
end procedure


----------------------------------------------------------------------
--  rat.e : based on rat v 2.40  (modified & trimmed)
----------------------------------------------------------------------

global constant
    lbp     = MOUSE_LEFTBUTTON,   -- left button pressed
    rbp     = MOUSE_RIGHTBUTTON,  -- right button pressed
    mbp     = MOUSE_MIDDLEBUTTON, -- middle button pressed
    rmoved  = 8     -- rat moved

global integer
    rat_event,      -- rat moved or a rat button is pressed
    rx,             -- current rat horizontal coordinate
    ry,             -- current rat vertical coordinate
    rat_visible     -- boolean, rat pointer visible or hidden (1 or 0)

integer maxrx, minrx, maxry, minry, hotx, hoty,
    lastx,         -- saved (last drawn) rat horizontal coordinate
    lasty          -- saved (last drawn) vertical coordinate

global procedure show_rat()
    place_mouse(rx,ry)
    mouse_pointer(1)
    rat_visible = 1
end procedure -- show_rat

global procedure hide_rat()
    place_mouse(rx,ry)
    mouse_pointer(0)
    rat_visible = 0
end procedure -- hide_rat

global procedure rat()
--
    sequence events
    integer rat_moved -- boolean, if true then the rat pointer
		      -- has moved since the last rat() call

    events = get_mouse()
    rat_event = events[1]
    rx = events[2]
    ry = events[3]

    rat_moved = 0
    if lastx != rx or lasty != ry then
      rat_moved = 1
    end if

    if rat_moved then
	rat_event += rmoved
	if rat_visible then
	    hide_rat()      -- erase image at previous location
	    lastx = rx
	    lasty = ry
	    show_rat()      -- show it again
	else
	    lastx = rx
	    lasty = ry
	end if
    end if
--
end procedure -- rat

global procedure confine_rat(
    integer x1, integer y1,     -- left top corner of screen cage
    integer x2, integer y2      -- right bottom corner
    )
    minrx = x1
    minry = y1
    maxrx = x2
    maxry = y2
    if rx<minrx then
	rx=minrx
    elsif rx>maxrx then
	rx=maxrx
    end if
    if ry<minry then
	ry=minry
    elsif ry>maxry then
	ry=maxry
    end if
end procedure -- confine_rat

global procedure set_rat_xy(integer x, integer y)
    -- places the rat pointer where you want it

    if x<minrx then
	x=minrx
    elsif x>maxrx then
	x=maxrx
    end if
    if y<minry then
	y=minry
    elsif y>maxry then
	y=maxry
    end if
    rx=x
    ry=y
    rat()
end procedure  -- set_rat_xy

global procedure rat_pointer(
    sequence p  -- rle pointer image with prepended hot spot coordinates
    )
--  select new pointer

    integer flag

    flag = rat_visible      -- save current rat visibility
    hide_rat()              -- make rat invisible
    hotx = p[1][1]-1
    hoty = p[1][2]-1
    set_mouse_cursor(p)
    if flag then show_rat() end if  -- make new pointer visible
end procedure -- rat_pointer

global procedure init_rat()
--  initialize - must be called *after* graphics mode is set
--  set pointer in middle of screen, hidden
--  confine rat to visible screen area

    sequence vc

    vc = video_config() -- machine_func(13,0)           -- video_config()
    rat_event = 0
    rx = floor(vc[5]/2)
    lastx = rx
    ry = floor(vc[6]/2)
    lasty = ry
    minrx = 0
    maxrx = vc[5]-1
    minry = 0
    maxry = vc[6]-1
end procedure -- init_rat

global procedure wait_while (integer pressed_button)
--  wait while specified rat button remains pressed

    while and_bits(rat_event, pressed_button) do
	rat()      -- keep rat pointer alive
    end while
end procedure -- wait_while

global function rat_zone(sequence zones)
--  scan zones for rat pointer presence
--  zones is sequence of rectangular zones to be tracked
--  each zone is defined by {x1, y1, x2, y2) sequence, where x1, y1
--  and x2, y2 are top left and  bottom right diagonal points resp.
--  return index of the *first* zone that includes the rat pointer,
--  or zero if none found

    sequence zi

    for i=1 to length(zones) do
	zi=zones[i]
	if rx>=zi[1] and
	   rx<=zi[3] and
	   ry>=zi[2] and
	   ry<=zi[4] then
	   return i
	end if
    end for
    return 0
end function -- rat_zone

----------------------------------------------------------------------
--  dragbox routines
----------------------------------------------------------------------

sequence dbox   -- drag box state: {x1,y1,x2,y2,color}

procedure safe_xor_box(integer x1, integer y1, integer x2, integer y2,
    integer color)
--  draw line box xor-ed with background
--  x1,y1 & x2,y2 are diagonal points
--  safe, because order of points is immaterial

    sequence s
    integer dummy, len

    if x1>x2 then   -- safety first
	dummy=x2
	x2=x1
	x1=dummy
    end if
    if y1>y2 then
	dummy=y2
	y2=y1
	y1=dummy
    end if
    len=x2-x1+1
    s=repeat(color,len)
    pixel(xor_bits(get_pixel({x1,y1,len}),s),{x1,y1})
    pixel(xor_bits(get_pixel({x1,y2,len}),s),{x1,y2})
    for y=y1+1 to y2-1 do
	pixel(xor_bits(get_pixel({x1,y}),color),{x1,y})
	pixel(xor_bits(get_pixel({x2,y}),color),{x2,y})
    end for
end procedure -- safe_xor_box

global procedure erase_drag_box()
--  erase drag box if one is drawn (visible)

    integer visible     -- rat visibility flag

    if dbox[5] then     -- box visible
	visible=rat_visible
	hide_rat()
	safe_xor_box(dbox[1],dbox[2],dbox[3],dbox[4],dbox[5])
	dbox[5]=0       -- box invisible
	if visible then show_rat() end if   -- restore rat
    end if
end procedure -- erase_drag_box

global function drag_box(
    integer color,              -- drag box color
    integer button_pressed
    )
    -- return drag box data: {x1,y1,x2,y2,color}
    -- this function integrates drag box with rat xor pointer:
    -- continually updates drag box data while specified rat
    -- button remains pressed
    -- color is drag box color (true only on BLACK background)

    integer x1, y1, x2, y2, visible

    visible = rat_visible

    if and_bits(button_pressed, rat_event) then
	erase_drag_box()
	x1=rx
	y1=ry
	x2=x1
	y2=y1
	safe_xor_box(x1,y1,x2,y2,color)
	while and_bits(rat_event, button_pressed) do
	    if rx!=x2 or ry!=y2 then
		hide_rat()
		safe_xor_box(x1,y1,x2,y2,color)
		x2=rx
		y2=ry
		safe_xor_box(x1,y1,x2,y2,color)
		if visible then
		    show_rat()
		end if
	    end if
	    rat()       -- keep rodent alive
	end while
	dbox = {x1,y1,x2,y2,color}
    end if
    return dbox
end function -- drag_box

-- init drag box data
dbox={0,0,0,0,0}    -- {x1,y1,x2,y2,color}
		    -- box not drawn if color=0


----------------------------------------------------------------------
--  various rat pointers
----------------------------------------------------------------------
global constant
RAT_ARROW = ARROW,
RAT_ARROW1 = ARROW1,
RAT_TRIANGLE = TRIANGLE,
RAT_CROSS = CROSSHAIR,
RAT_IBEAM = IBEAM,
RAT_LTRB = LTRB,
RAT_RTLB = RTLB,
RAT_LR = LR,
RAT_TB = TB,
RAT_PENCIL = PENCIL,
RAT_HGLASS = HGLASS,
RAT_NOGO = NOGO,
RAT_RESIZE = RESIZE,
RAT_HAND = HAND,
RAT_HAND1 = HAND1,
RAT_HAND2 = HAND2

-- init rat ----------------------------------------------------------
rat_visible = 0         -- let's make rat initially hidden
init_rat()
rat_pointer(RAT_ARROW)

-----------------------------------------------------------------
--  list.e  version 2.00  (trimmed)
-----------------------------------------------------------------

global function index(sequence list, object key)
--  return *key* index, or zero if key not found
    return find(key, list[1])
end function

global function fetch(sequence list, object key)
--  return list item with indicated key
    integer i

    i = find(key, list[1])
    if i then
	return list[2][i]
    end if
    crash_message("fetch error: key \"" & key & "\" not found...\n")
end function -- fetch

global function Fetch(
    object list,        -- list from which value is to be fetched
    sequence keys)      -- sequence of list keys, one for each level
			-- last one directly associated with fetched
			-- value
--  return item from *nested* list : 'deep' fetch

    for i=1 to length(keys) do
	list = fetch(list, keys[i])
    end for
    return list
end function -- Fetch

global function set(sequence list, object key, object new_value)
--  replace specified list item with a new value
--  return modified list
    integer i

    i = find(key, list[1])      -- key index
    if i then
	list[2][i] = new_value
	return list
    end if
    crash_message("store error: key \"" & key & "\" not found...\n")
end function -- set

global function Set(sequence list, sequence keys, object new_value)
--  'deep', recursive set for *nested* lists
    integer i,len

    len = length(keys)
    i = find(keys[1],list[1])
    if i then
	if len > 1 then
	    list[2][i] = Set(list[2][i], keys[2..len], new_value)
	else
	    list[2][i] = new_value
	end if
	return list
    end if
    crash_message("Set error: key \"" & keys[1] & "\" not found...\n")
end function -- Set

global function add(sequence list, object key, object val)
--  append list
    return {append(list[1],key), append(list[2],val)}
end function -- add

global function join(sequence list1, sequence list2)
--  concatenate list1 and list2
    return {list1[1] & list2[1],list1[2] & list2[2]}
end function

global function delete(sequence list, object key)
--  delete item with the specified key from the list
    integer i,l

    i = find(key, list[1])      -- index
    if i then
	l = length(list[1])
	return {list[1][1..i-1] & list[1][i+1..l],
	    list[2][1..i-1] & list[2][i+1..l]}
    end if
    crash_message("delete error: key \"" & key & "\" not found...\n")
end function -- delete


----------------------------------------------------------------------
-- widgets : initialization & general code
----------------------------------------------------------------------

global constant
    kbd_event = 16,             -- keyboard event
    rat_events = 15,            -- all rat events combined (mask)
    rgb={52,52,52}              -- default primary color palette (color 8)

global sequence wid,pile,widgets
global integer exitf,pressed_key,hsize,vsize,menuf,pilef,pilep,screen,self
sequence clicked, keyed
integer emask                   -- event mask

global procedure set_default_colors(
    sequence mix -- mix = {red,green,blue} primary face color mix
    )
    -- set graduated scale of standard colors (0,3,7,8,11,15)

    object p
    p = {
	{0,0,0},            -- black, text
	{0,0,42},           -- blue
	{0,42,0},           -- green
	floor(mix*0.50),    -- shadows
	{42,0,0},           -- red
	{42,0,42},          -- magenta
	{42,21,0},          -- brown
	floor(mix*0.85),    -- face
	mix,                -- light face ('paper'color)
	{21,21,63},         -- light blue
	{21,63,21},         -- light green
	floor((mix+64)/2),  -- highlights
	{63,21,21},         -- light red
	{63,21,63},         -- light magenta
	{63,63,21},         -- yellow
	{63,63,63}}         -- white
    set_colors(0, p)
end procedure -- set_default_colors

global procedure set_mode(integer mode)
    --  invoke desired graphics mode and set standard colors

    object vc

    if graphics_mode(mode) then
	puts(1,"Requested graphics mode not available!\n")
    end if
    set_default_colors(rgb)
    vc = video_config()
    hsize = vc[VC_XPIXELS]
    vsize = vc[VC_YPIXELS]
    init_rat()
end procedure -- set_mode

global procedure set_text_mode()
integer ok

    if graphics_mode(-1) then end if
    for i = 0 to 15 do
      ok = vga_setegacolor(i)
    end for

end procedure -- set_text_mode

global procedure zap(integer w)
    -- zap widget and all its components to recover some memory

    sequence c,s

    s = wid[w]
    if index(s,"par") then
	c = fetch(s,"par")
	for i=1 to length(c[1]) do
	    zap(c[2][i])
	end for
    end if
    wid[w] = {}
end procedure -- zap

global procedure shift(integer w, integer dx, integer dy)
    -- translate widget and all its components by indicated
    -- number of pixels in both directions

    sequence c,s
    integer j,k,m

    s = wid[w]
    if index(s,"par") then
	c = fetch(s,"par")
	for i=1 to length(c[1]) do
	    shift(c[2][i], dx, dy)
	end for
    end if
    j = index(s, "x")
    if j then
	s[2][j] += dx
    end if
    k = index(s, "y")
    if k then
	s[2][k] += dy
    end if
    m = index(s,"box")
    if m then
	s[2][m][1] += dx
	s[2][m][2] += dy
    end if
    if j or k or m then
	wid[w] = s
    end if
end procedure -- shift


----------------------------------------------------------------------
--  items (properties & methods) and components
----------------------------------------------------------------------

global function item(       -- fetch named item of widget
    integer wi,             -- widget (index)
    object name             -- item name, key
    )
    return fetch(wid[wi], name)
end function -- item

global procedure bind(      -- reset existing item to new value
    integer wi,             -- widget (index)
    object name,            -- item name, key
    object val              -- new value
    )
    wid[wi][2][find(name, wid[wi][1])] = val
end procedure -- bind

global procedure make(
    integer wi,             -- widget (index)
    object name,            -- item name, key
    object val              -- new value
    )
    -- add new item to list if it does not exist,
    -- otherwise just reset its value

    integer i

    i = index(wid[wi], name)
    if i then
	wid[wi][2][i] = val
    else
	wid[wi][1] = append(wid[wi][1], name)
	wid[wi][2] = append(wid[wi][2], val)
    end if
end procedure -- make

global function part(   -- return named component (index)
    integer wi,         -- widget (index)
    object name         -- part name
    )
    return Fetch(wid[wi], {"par", name})
end function -- part

global procedure affix( -- add new part
    integer wm,         -- 'boss' widget (index)
    object name,        -- part name, key
    integer wi          -- (new) part stack index
    )
    integer i

    i = index(wid[wm], "par")
    if i then           -- parts already exit, append them
	wid[wm][2][i][1] = append(wid[wm][2][i][1], name)
	wid[wm][2][i][2] = append(wid[wm][2][i][2], wi)
    else                -- parts do not exit, generate them
	make(wm, "par", {{name},{wi}})
    end if
    make(wi, "mai", wm)
end procedure -- affix

global function main(integer wi)
    -- return 'main' widget (index)

    integer i

    i = find("mai", wid[wi][1])
    if i then
	return wid[wi][2][i]
    end if
    return 0
end function


----------------------------------------------------------------------
-- 'generic' methods
----------------------------------------------------------------------

global procedure show(object wi)
    if atom(wi) then
	call_proc(item(wi, "sho"), {wi})
    else
	for i=1 to length(wi) do
	    call_proc(item(wi[i], "sho"), {wi[i]})
	end for
    end if
end procedure -- show

global procedure hide(object wi)
    if atom(wi) then
	call_proc(item(wi, "hid"), {wi})
    else
	for i=1 to length(wi) do
	    call_proc(item(wi[i], "hid"), {wi[i]})
	end for
    end if
end procedure -- hide

global procedure on_rat()
    sequence s

    s = item(self, "rat")
    for i=1 to length(s) do
	call_proc(s[i], {})
    end for
end procedure -- on_rat

global procedure on_key()
    sequence s

    s = item(self, "kbd")
    for i=1 to length(s) do
	call_proc(s[i], {})
    end for
end procedure -- on_key

global procedure arat(integer wi, integer rid)
-- append rid to widget's 'rat' sequence of routine ids

    sequence s

    s = item(wi, "rat")
    bind(wi, "rat", s&rid)
end procedure -- arat

global procedure akbd(integer wi, integer rid)
-- append rid to widget's 'kbd' sequence of routine ids

    sequence s

    s = item(wi, "kbd")
    bind(wi, "kbd", s&rid)
end procedure -- akbd

global procedure splice(integer wi, integer rid)
-- append rid to both 'rat' and 'kbd sequences of routine ids

    sequence s

    s = item(wi, "rat")
    bind(wi, "rat", s&rid)
    s = item(wi, "kbd")
    bind(wi, "kbd", s&rid)
end procedure -- splice

----------------------------------------------------------------------
-- zones & tracking
----------------------------------------------------------------------

procedure Track(integer w)
-- get all rat zones traps and key traps in widget w

    sequence k,p,s,u
    integer e,i,l,x,y

    p = wid[w]
    i = find("eve",p[1])          -- index
    if i then
	e = p[2][i]
	if and_bits(e, kbd_event) then
	    k = fetch(p, "key")
	    l = length(k)
	    if l then
		for j=1 to l do
		    keyed = append(keyed, {w,k[j]})
		end for
		emask = or_bits(emask, kbd_event) -- events to be tracked
	    end if
	end if
	e = and_bits(e, rat_events)
	if e then
	    u = item(w, "box")
	    x = u[1]
	    y = u[2]
	    clicked = append(clicked, {w , e, x, y, x+u[3]-1, y+u[4]-1})
	    emask = or_bits(emask,e)    -- events to be tracked
	end if
    end if
    i = find("par", p[1])          -- index
    if i then
	s = p[2][i][2]  -- part sequence
	for j=1 to length(s) do
	    Track(s[j])
	end for
    end if
end procedure -- Track

global procedure track(object ws)
    -- collect all traps in all widgets in sequence ws, recursively,
    -- into easily 'scannable' structures:
    --  "clicked" : {{widget_index, event, x1, y1, x2, y2}, {...
    --  "keyed" : {{widget_index, key}, {...,
    --  where key is ascii code

    emask = 0       -- events to be tracked
    clicked = {}
    keyed = {}
    for i=1 to length(ws) do
	Track(ws[i])
    end for
end procedure -- track

global function get_event(integer mask)
    -- update rat and wait for specified event
    -- return event code

    integer e

    while 1 do
	rat()
	e = and_bits(rat_event, mask)       -- check rat first
	if e then
	    return e
	end if
	pressed_key = get_key()
	if pressed_key != -1 then
	    if and_bits(mask, kbd_event) or pressed_key = esc then
		return kbd_event
	    end if
	end if
    end while
end function -- get_event

global procedure collapse_menu()
    integer l,wi

    l = length(pile)
    while l > pilep do
	wi = pile[l]        -- submenu on top of pile
	hide(wi)
	l -= 1
	pile = pile[1..l]
    end while
    menuf = false
end procedure -- collapse_menu

global procedure loop()
    -- track widgets, including their parts and
    -- respond to *all* events as they occur

    sequence zi
    integer event,flag,len

    exitf = false
    if screen then
	widgets &= screen
    end if
    show_rat()
    while not exitf do
	len = length(pile)
	if len then             -- window or submenu in use
	    track({pile[len]})  -- track the topmost item
	else                    -- no more menus nor windows
	    track(widgets)
	end if
	pilef = false
	while not exitf and not pilef do
	    event = get_event(emask)
	    if event = kbd_event then
		flag = false
		for i=1 to length(keyed) do
		    if keyed[i][2] = pressed_key then
			self = keyed[i][1]
			on_key()
			flag = true
			exit                -- exit for loop
		    end if
		end for
		if flag = false and pressed_key = esc then
		    exitf = true
		end if
	    end if
	    if and_bits(event, lbp)  then
		flag = false
		for i=1 to length(clicked) do
		    zi = clicked[i]
		    if and_bits(zi[2], lbp) and
			rx>=zi[3] and ry>=zi[4] and rx<=zi[5]
			and ry<=zi[6] then
			    flag = true
			    self = zi[1]
			    on_rat()
			    exit            -- exit for loop
		    end if
		end for
		if not flag and menuf then  -- click outside: collapse menus
		    collapse_menu()
		end if
	    end if
	end while
    end while
end procedure -- loop


----------------------------------------------------------------------
-- text : built-in text widget
----------------------------------------------------------------------

include sysb13.e    -- font

global integer default_font
integer m

default_font = sysb13

global function text(
    integer x,
    integer y,
    sequence s
    )
    wid = append(wid, {{"x","y","str","sho","fon","for","hil","sha","pap","att",
	"jus","ecp","elp","hso","vso"},{x,y,s,m,default_font,BLACK,BRIGHT_RED,
	3,7, HILITE, LEFT, 0, 1, 1, 1}})
    return length(wid)
end function -- text

global procedure retext(        -- new word: REset TEXT environment
    integer wi                  -- text widget
    )
    sequence p

    p = wid[wi][2]              -- values sequence of text list
    select_font(p[5])
    set_foreground(p[6])
    set_hilite_color(p[7])
    set_shadow_color(p[8])
    set_paper_color(p[9])
    set_attributes(p[10])
    justify(p[11])
    set_extra_char_pitch(p[12])
    set_extra_line_pitch(p[13])
    set_shadow_offsets(p[14], p[15])
end procedure -- retext

global procedure show_text(
    integer wi                  -- text widget
    )
    sequence p

    p = wid[wi][2]              -- values sequence of text list
    retext(wi)
    setxy(p[1], p[2])
    hide_rat()
    write(p[3])
    show_rat()
end procedure -- show_text
m = routine_id("show_text")


----------------------------------------------------------------------
-- activate / deactivate : auxiliary global routines
----------------------------------------------------------------------

constant dimmed = {{3,7},{7,3}}

global procedure activate(
    integer wi,         -- (re)activate this widget
    sequence name       -- name of its associated text widget
    )
    integer wt

    if not find(wi, widgets) then
	widgets &= wi
    end if
    wt = part(wi, name)
    bind(wt, "for", BLACK)
    bind(wt, "hil", BRIGHT_RED)
    bind(wt, "att", HILITE)
    show(wi)
    track(widgets)
end procedure -- activate

global procedure deactivate(
    integer wi,         -- deactivate this widget
    sequence name       -- name of its associated text widget
    )
    integer i,wt

    i = find(wi, widgets)
    if i then                   -- remove it from tracked bunch
	widgets = widgets[1..i-1] & widgets[i+1..length(widgets)]
    end if
    wt = part(wi, name)
    bind(wt, "for", dimmed)
    bind(wt, "hil", dimmed)
    show(wi)
    track(widgets)
end procedure -- deactivate


----------------------------------------------------------------------
-- initialize widgets core
----------------------------------------------------------------------

screen = 0      -- default
widgets = {}    -- trackable widgets (default)
menuf = 0       -- no active menu
wid = {}        -- stack of widgets
pile = {}       -- windows & menus pile
pilep = 0       -- pile pointer : points to top of pile just before
		-- current menu or menus were added
--tick_rate(200)


----------------------------------------------------------------------
--  Pete's smart_print -  mods by Gabriel Boehme & me
----------------------------------------------------------------------

function smart_print_string(sequence s)
--  faster, I think, as a function

    object x

    for i = 1 to length(s) do
	x = s[i]
	if not integer(x) or x < ' ' or x > 126 or x = '"' then
	    return 0
	end if
    end for
    return 1
end function

global procedure sp(object o)  -- smart print
    if atom(o) then
	printf(1, "%g", {o})
    elsif length(o) = 0 then
	puts(1, "{}")
    elsif smart_print_string(o) then
	printf(1, "\"%s\"", {o})
    else
	puts(1, "{")
	sp(o[1])
	for i = 2 to length(o) do
	    puts(1, ",")
	    sp(o[i])
	end for
	puts(1, "}")
    end if
end procedure

global procedure spn(object o) -- smart print & newline
    sp(o)
    puts(1, "\n")
end procedure


----------------------------------------------------------------------
-- debugging
----------------------------------------------------------------------

global procedure pw(object w)
--  formatted print of widget sequence

    if atom(w) then w = {w} end if
    for i=1 to length(w) do
	printf(1, "widget %d:\n",w[i])
	puts(1,"names: ")
	spn(wid[w[i]][1])
	puts(1,"items: ")
	spn(wid[w[i]][2])
	puts(1,"\n")
    end for
end procedure

